
/**************************************************************************
 *
 *  FILE            mrwalker.c
 *  MODULE OF       
 *
 *  DESCRIPTION     
 *
 *  WRITTEN BY      Sverre H. Huseby
 *
 *  MODIFICATIONS   
 *
 **************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/time.h>
#include <time.h>
#include <unistd.h>
#include <signal.h>
#include <math.h> /* random, srandom */

#include <shhmsg.h>
#include <shhnet.h>
#include <xalloc.h>

#include "viking.h"

#define PLAYER_NAME  "MrWalker"
/* the following include-file defines PLAYER_PASS as the password string */
#include "password.h"

#define CONV_LOG     "converse.log"
#define PAL_FILE     "pals.txt"

#define GREET_LIMIT  20

/**************************************************************************
 *                                                                        *
 *                       P R I V A T E    D A T A                         *
 *                                                                        *
 **************************************************************************/

static char **adverb;
static char **feeling;
static char **saying;
static char **repTell;
static char **repReply;
static char **repAsk;
static char **repYesNo;
static char **repRobot;
static char **repSpam;
static char **repBad;
static int  nAdverbs;
static int  nFeelings;
static int  nSayings;
static int  nRepTell;
static int  nRepReply;
static int  nRepAsk;
static int  nRepYesNo;
static int  nRepRobot;
static int  nRepSpam;
static int  nRepBad;

static struct opposite {
    char *s1, *s2;
} opposite[] = {
    { "n",      "s"   },
    { "e",      "w"   },
    { "se",     "nw"  },
    { "sw",     "ne"  },
    { "u",      "d"   },
    { "enter",  "out" },
};
char *direction[] = {
    "n", "s", "e", "w",
    "se", "nw", "sw", "ne",
    "u", "d",
    "enter", "in", "out"
};
#define NUM_DIRECTIONS (sizeof(direction) / sizeof(char *))

static int mustRestart = 0;
static int fleeing = 0;

#define MAX_PALS 300
typedef struct pal_tag {
    char   *name;
    char   *real;  /* NULL if unknown */
    int    n;
    time_t when;
} Pal;
static Pal pal[MAX_PALS];
static int nPals = 0;

static char prevExit[20] = "";

static int hp, sp, xp, xpnl, cash;

/* stay away from areas made by these creators */
static char **madCreator;
static int  nMadCreator;
static char prevCreator[80];

static volatile int mustReread = 0;



/**************************************************************************
 *                                                                        *
 *                   P R I V A T E    F U N C T I O N S                   *
 *                                                                        *
 **************************************************************************/

static char *stristr(const char *haystack, const char *needle)
{
    const char *n, *h;
    
    while (*haystack) {
        h = haystack;
        n = needle;
        while (*n) {
            if (tolower(*n) != tolower(*h))
                break;
            ++n;
            ++h;
        }
        if (!*n)
            return (char *) haystack;
        ++haystack;
    }
    return NULL;
}

static int stricmp(const char *s1, const char *s2)
{
    int diff;
    
    do {
	diff = tolower(*s2++) - tolower(*s1);
    } while (!diff && *s1++);
    return diff;
}

static void sigTermHandler(int s)
{
    vikLogout();
    exit(0);
}

static void sigHupHandler(int s)
{
    msgVerbose(1, "***  got SIGHUP...\n");
    mustReread = 1;
}

static void sigLinkdeadHandler(int s)
{
    msgVerbose(1, "***  got SIGUSR1... going linkdead\n");
    exit(0);
}

static void setupSigHandlers(void)
{
    struct sigaction sact;

    sigemptyset(&sact.sa_mask);
  #ifdef SA_RESTART
    sact.sa_flags = SA_RESTART;
  #else
    sact.sa_flags = 0;
  #endif

    sact.sa_handler = sigTermHandler;
    sigaction(SIGINT, &sact, NULL);
    sigaction(SIGTERM, &sact, NULL);
    sigaction(SIGQUIT, &sact, NULL);
    sigaction(SIGABRT, &sact, NULL);

    sact.sa_handler = sigHupHandler;
    sigaction(SIGHUP, &sact, NULL);

    sact.sa_handler = sigLinkdeadHandler;
    sigaction(SIGUSR1, &sact, NULL);
}

static void freePals(void)
{
    int q;

    for (q = 0; q < nPals; q++) {
	free(pal[q].name);
	free(pal[q].name);
    }
    nPals = 0;
}

static void readPals(void)
{
    int  i;
    FILE *f;
    char line[200], *ret, *p, *tok;

    freePals();
    if ((f = fopen(PAL_FILE, "rt")) == NULL)
	return;
    nPals = 0;
    for (;;) {
	while ((ret = fgets(line, sizeof(line), f)) != NULL) {
	    if ((p = strchr(line, '#')) != NULL)
		*p = '\0';
	    while (*line && isspace(line[i = strlen(line) - 1]))
		line[i] = '\0';
	    if (*line)
		break;
	}
	if (!ret)
	    break;
	if (nPals >= MAX_PALS)
	    break;
        if ((tok = strtok(line, " \t")) == NULL)
	    msgFatal("error in " PAL_FILE "\n");
	pal[nPals].name = xstrdup(tok);
        if ((tok = strtok(NULL, " \t")) == NULL)
	    msgFatal("error in " PAL_FILE "\n");
	pal[nPals].n = atoi(tok);
        if ((tok = strtok(NULL, " \t")) == NULL)
	    msgFatal("error in " PAL_FILE "\n");
	pal[nPals].when = (time_t) atol(tok);
        if ((tok = strtok(NULL, " \t")) == NULL)
	    pal[nPals].real = NULL;
	else
	    pal[nPals].real = xstrdup(tok);
        ++nPals;
    }
    fclose(f);
    msgVerbose(1, "***  read %d pals from \"%s\"\n", nPals, PAL_FILE);
}

static void writePals(void)
{
    int  q;
    FILE *f;
    
    if ((f = fopen(PAL_FILE, "wt")) == NULL)
	msgFatal("couldn't write " PAL_FILE "\n");
    for (q = 0; q < nPals; q++) {
	fprintf(f, "%-10s %3d %10ld%s%s\n",
		pal[q].name, pal[q].n, (long) pal[q].when,
		pal[q].real ? " " : "",
		pal[q].real ? pal[q].real : "");
    }
    fclose(f);
}

static Pal *findPal(const char *name)
{
    int q;

    for (q = 0; q < nPals; q++)
	if (stricmp(pal[q].name, name) == 0)
	    return &pal[q];
    return NULL;
}

static Pal *updatePal(const char *name)
{
    Pal *ret;

    if ((ret = findPal(name)) == NULL) {
	if (nPals >= MAX_PALS)
	    return NULL;
	ret = &pal[nPals];
	ret->name = xstrdup(name);
	ret->n = 0;
	ret->real = NULL;
	++nPals;
    }
    ++ret->n;
    ret->when = time(NULL);
    return ret;
}

static void setPalReal(const char *name, const char *real)
{
    Pal *p;

    if ((p = findPal(name)) == NULL)
	return;
    free(p->real);
    if (real && strlen(real))
	p->real = xstrdup(real);
    else
	p->real = NULL;
}

static void readTextFiles(void)
{
    txtFree(madCreator, nMadCreator);
    madCreator = txtReadWords("madcreator.txt", &nMadCreator);

    txtFree(adverb, nAdverbs);
    adverb     = txtReadWords("adverbs.txt", &nAdverbs);
    txtFree(feeling, nFeelings);
    feeling    = txtReadWords("feelings.txt", &nFeelings); 

    txtFree(saying, nSayings);
    saying     = txtReadLines("sayings.txt", &nSayings);
    txtFree(repTell, nRepTell);
    repTell    = txtReadLines("reptell.txt", &nRepTell);
    txtFree(repReply, nRepReply);
    repReply   = txtReadLines("repreply.txt", &nRepReply);
    txtFree(repAsk, nRepAsk);
    repAsk     = txtReadLines("repask.txt", &nRepAsk);
    txtFree(repYesNo, nRepYesNo);
    repYesNo   = txtReadLines("repyesno.txt", &nRepYesNo);
    txtFree(repRobot, nRepRobot);
    repRobot   = txtReadLines("reprobot.txt", &nRepRobot);
    txtFree(repSpam, nRepSpam);
    repSpam    = txtReadLines("repspam.txt", &nRepSpam);
    txtFree(repBad, nRepBad);
    repBad     = txtReadLines("repbad.txt", &nRepBad);
    readPals();
    mustReread = 0;
}

static char *oppositeDirection(const char *dir)
{
    int q;

    for (q = 0; q < sizeof(opposite) / sizeof(struct opposite); q++) {
        if (strcmp(dir, opposite[q].s1) == 0)
            return opposite[q].s2;
        if (strcmp(dir, opposite[q].s2) == 0)
            return opposite[q].s1;
    }
    return "?";
}

static void reply(char *line, char **list, int n)
{
    char who[200], first[200], greet[200], *reply, *whoend, *p, *p2, *name;
    Pal *pl;
    static time_t lastSpamReply = 0;
    int dontReply = 0;

    while (*line && (isspace(*line) || *line == '>'))
	++line;
    if ((p = strstr(line, "you:")) == NULL)
	return;
    p2 = p - 1;
    while (p2 >= line && isspace(*p2))
	--p2;
    while (p2 >= line && !isspace(*p2))
	--p2;
    while (p2 >= line && isspace(*p2))
	--p2;
    while (p2 >= line && !isspace(*p2))
	--p2;
    line = p2 + 1;
    p += 4;
    while (*p && isspace(*p))
	++p;
    if (!*p || strlen(p) < 3)
	return;
    /* first check if someone is talking about robots */
    if (stristr(p, "robot") || stristr(p, "machine") || stristr(p, "human")
	|| stristr(p, " bot ") || stristr(p, "cyborg")
	|| stristr(p, "script")) {
	reply = txtRandom(repRobot, nRepRobot);
    } else if (stristr(p, "fuck") || stristr(p, "shit")
	       || stristr(p, "asshole") || stristr(p, "damn")
	       || stristr(p, "suck") || stristr(p, " piss ")
	       || stristr(p, " hell ")) {
	reply = txtRandom(repBad, nRepBad);
    } else if (stristr(p, "sorry") || stristr(p, "excuse")) {
	reply = "Ok.";
    } else {
	if (strchr(p, '?')) {
	    list = repAsk;
	    n = nRepTell;
	}
	strcpy(first, p);
	if ((p2 = strchr(first, ' ')) != NULL)
	    *p2 = '\0';
	if (stricmp(first, "is") == 0 || stricmp(first, "isn't") == 0
	    || stricmp(first, "are") == 0 || stricmp(first, "aren't") == 0
	    || stricmp(first, "were") == 0 || stricmp(first, "weren't") == 0
	    || stricmp(first, "was") == 0 || stricmp(first, "wasn't") == 0
	    || stricmp(first, "have") == 0 || stricmp(first, "haven't") == 0
	    || stricmp(first, "can") == 0 || stricmp(first, "can't") == 0
	    || stricmp(first, "could") == 0 || stricmp(first, "couldn't") == 0
	    || stricmp(first, "had") == 0 || stricmp(first, "hadn't") == 0
	    || stricmp(first, "did") == 0 || stricmp(first, "didn't") == 0) {
	    list = repYesNo;
	    n = nRepYesNo;
	}
	reply = txtRandom(list, n);
    }
    strncpy(who, line, 80);
    if ((whoend = strchr(who, ' ')) != NULL)
	*whoend = '\0';
    *greet = '\0';
    if ((pl = findPal(who)) != NULL) {
        time_t diff, now;
        char   buff[200];

	if (pl->real)
	    name = pl->real;
	else
	    name = pl->name;
	now = time(NULL);
        diff = now - pl->when;
        if (diff <= 1) {
	    if (now >= lastSpamReply + 4) {
		reply = txtRandom(repSpam, nRepSpam);
		lastSpamReply = now;
	    } else
		dontReply = 1;
        } else if (diff > 2 * 24 * 60 * 60) {
	    sprintf(buff, "Hi, %s! Long time no see...", name);
	    reply = buff;
        } else if (diff > 1 * 24 * 60 * 60) {
            sprintf(buff, "Hi, %s. Haven't heard from you in a day or so...",
		    name);
            reply = buff;
        }
	if ((random() % 5) == 0) {
	    switch (random() % 4) {
	      case 0:
		strcpy(greet, "Well, ");
		break;
	      case 1:
		strcpy(greet, "Ok, ");
		break;
	      case 2:
		strcpy(greet, "");
		break;
	      default:
		*greet = '\0';
	    }
	    strcat(greet, name);
	    strcat(greet, "; ");
	}
    }
    updatePal(who);
    writePals();
    if (!dontReply) {
	vikLog(CONV_LOG, "%s\nreply %s %s%s\n", line, who, greet, reply);
	vikOutQueue(4, "reply %s %s%s", who, greet, reply);
    }
}
		  
static void respTell(char *line)
{
    reply(line, repTell, nRepTell);
}

static void respReply(char *line)
{
    reply(line, repReply, nRepReply);
}

static void respAsk(char *line)
{
    reply(line, repAsk, nRepAsk);
}

static void hookPray(char *line)
{
    vikOut("pray");
}

static void hookRestart(char *line)
{
    switch (random() % 3) {
      case 0:
	vikOut("chat Hurray! I made it! I died again!");
	break;
      case 1:
	vikOut("chat That's it. Dead again.");
	break;
      default:
	vikOut("chat Ahh! Once more I get to meet Odin...");
    }	
    mustRestart = 1;
}

static void hookFight(char *line)
{
    char *p;
    int  q;
    
    if (hp >= 60 && (p = strchr(line, ' ')) != NULL) {
	q = (random() % 4);
	fleeing = 0;
	while (hp >= 60 && q-- && !fleeing) {
	    *p = '\0';
	    vikOut("kick %s", line);
	    vikOut("punch %s", line);
	    vikSleep(1);
	}
    }
    vikOut("flee");
    *prevExit = '\0';
}

static void hookKill(char *line)
{
    char *who, *whoend;

    if (hp < 40) {
	vikOut("flee");
	*prevExit = '\0';
	return;
    }
    if ((who = strstr(line, "You notice ")) == NULL)
        return;
    who += strlen("You notice ");
    if ((whoend = strchr(who, ' ')) == NULL)
        return;
    *whoend = '\0';
    vikOut("kill %s", who);
}

static void hookFleeing(char *line)
{
    fleeing = 1;
}

static void hookShop(char *line)
{
    vikOut("untag all");
    vikOut("tag knife");
    vikOut("tag sword");
    vikOut("unequip");
    vikOut("sell all");
    vikOut("drop all");
    vikSleep(4);
    if (cash > 0) {
	vikOut("untag all");
	vikOut("buy sword");
	vikOut("buy knife");
	vikOut("buy platemail");
	vikOut("wield all");
	vikOut("wear all");
    }
    strcpy(prevExit, "s");
    vikOut(prevExit);
}

static void hookBank(char *line)
{
    if (cash > 500)
	vikOut("deposit %d", cash - 500);
    else
	vikOut("withdraw all");
}

static void hookPub(char *line)
{
    vikOut("say play b1");
    vikOut("say play b2");
    vikOut("say play d3");
}

static void hookAdvance(char *line)
{
    int q;

    for (q = 0; q < 1; q++) {
	vikOut("advance str");
	vikOut("advance dex");
	vikOut("advance int");
	vikOut("advance con");
    }
    if (xpnl <= 0)
	vikOut("advance");
}

static void hookAnarchVote(char *line)
{
    vikOutQueue(2, "no");
}

static void hookDoor(char *line)
{
    vikOut("open door");
}

static void hookStatus(char *line)
{
    char *p, *txt = "(STATUS) ";
    int  hp2, sp2, xp2, xpnl2, cash2;

    if ((p = strstr(line, txt)) == NULL)
	return;
    p += strlen(txt);
    if (sscanf(p, "H:%d S:%d X:%d XL:%d C:%d",
	       &hp2, &sp2, &xp2, &xpnl2, &cash2) == 5) {
	hp = hp2;
	sp = sp2;
	xp = xp2;
	xpnl = xpnl2;
	cash = cash2;
    }
}

static void hookEntersGame(char *line)
{
    char who[200], greet[200], *p, *p2;
    Pal *pl;

    while (*line && (isspace(*line) || *line == '>'))
	++line;
    if ((p = strstr(line, "[INFO]:")) == NULL)
	return;
    p += 7;
    while (*p && isspace(*p))
	++p;
    p2 = p;
    while (*p2 && !isspace(*p2))
	++p2;
    *p2 = '\0';
    strncpy(who, p, 80);
    *who = tolower(*who);
    msgVerbose(1, "*** \"%s\" entered the game.\n", who);
    *greet = '\0';
    if ((pl = findPal(who)) != NULL && pl->n > GREET_LIMIT
	&& (random() % 2) == 0) {
	strcpy(greet, "chat feeling: ");
	switch (random() % 3) {
	  case 0:
	    strcat(greet, "bow ");
	    break;
	  case 1:
	    strcat(greet, "dance ");
	    break;
	  case 2:
	    strcat(greet, "ckiss ");
	    break;
	  default:
	    *greet = '\0';
	}
	strcat(greet, who);
	vikOutQueue(4, "%s", greet);
    } else {
	if (!pl)
	    msgVerbose(1, "*** no greeting done: not my pal.\n");
	else if (pl->n <= GREET_LIMIT)
	    msgVerbose(1, "*** no greeting done: not pal enough.\n");
	else
	    msgVerbose(1, "*** no greeting done.\n");
    }
}

/* 0 if nothing special, 1 if new direction already choosen. */
static int checkCreator(void)
{
    int  q, ret = 0, danger;
    char *creator, *dir, line[200];
    
    vikOut("creator");
    if (vikWaitForTimeout("Your delicate mind senses "
			  "that this", 3000) == 0) {
	vikIn(line, sizeof(line));
	if ((creator = strchr(line, '\'')) == NULL)
	    msgVerbose(1, "*** Strange: no \' in creator line\n");
	else {
	    *creator = '\0';
	    while (!isspace(*(creator - 1)))
		--creator;
	    msgVerbose(2, "***  creator: \"%s\"\n", creator);
	    danger = 0;
	    for (q = 0; q < sizeof(madCreator) / sizeof(char *); q++) {
		if (stricmp(creator, madCreator[q]) == 0) {
		    msgVerbose(1, "***  madCreator "
			       "%s made this area.\n", creator);
		    danger = 1;
		    break;
		}
	    }
	    if (danger && stricmp(prevCreator, creator) != 0) {
		dir = oppositeDirection(prevExit);
		if (isalpha(*dir)) {
		    msgVerbose(1, "***  returning %s\n", dir);
		    vikOut(dir);
		    strcpy(prevExit, dir);
		    ret = 1;
		} else
		    msgVerbose(1, "***  Oops! Don't know how to turn!\n");
	    }
	    strncpy(prevCreator, creator, sizeof(prevCreator) - 1);
	}
    }
    return ret;
}

static void robot(void)
{
    char   line[200], *s, *tok, str[80];
    char   *exits[10];
    int    nExit, n;
    int    invWhen = 0, wearWhen = 0, scoreWhen = 0, timeWhen = 0, eatWhen = 0;
    time_t start;

    start = time(NULL);
    vikOut("wimpy 40");
    vikOut("screen width 195");
    vikOut("party add hugnamordir");
    vikOut("prompt (STATUS) H:#H S:#S X:#X XL:#XL C:#C>");
    vikOut("pray");
    vikAddHook(" tells you:", respTell, 1);
    vikAddHook(" replys you:", respReply, 1);
    vikAddHook(" asks you:", respAsk, 1);
    vikAddHook(" shop", hookShop, 1);
    vikAddHook(" Shop", hookShop, 1);
    vikAddHook(" The Bank.", hookBank, 1);
    vikAddHook(" local pub.", hookPub, 1);
    vikAddHook(" hit you", hookFight, 1);
    vikAddHook(" tickled you in the", hookFight, 1);
    vikAddHook(" approaching you with murder in ", hookKill, 1);
    vikAddHook("It seems vaguely familiar...", hookPray, 1);
    vikAddHook("The altar to Odin.", hookPray, 1);
    vikAddHook("your own dead body from above.", hookRestart, 1);
    vikAddHook("The adventurers guild.", hookAdvance, 1);
    vikAddHook("Your legs run away with you!", hookFleeing, 0);
    vikAddHook("ANARCHY VOTE HAS STARTED.", hookAnarchVote, 1);
    vikAddHook("(STATUS)", hookStatus, 0);
    vikAddHook("bump into a", hookDoor, 0);
    vikAddHook("enters the game.", hookEntersGame, 1);
    vikSleep(5);
    /* go to the pub, hoping to get some xp. */
    vikOut("s"); vikOut("e"); vikOut("e"); vikOut("n"); vikOut("e");
    vikSleep(5);
    /* go to the adventurer's guild, in case we can increase level. */
    vikOut("w"); vikOut("s"); vikOut("e"); vikOut("s");
    vikSleep(5);
    /* go to the shop */
    vikOut("n"); vikOut("n");
    while (time(NULL) < start + (48 * 60 * 60L)) {
        if (mustRestart) {
            mustRestart = 0;
            vikLogout();
            sleep(30);
            vikLogin(PLAYER_NAME, PLAYER_PASS);
            vikSleep(5);
        }
	if (mustReread) {
	    readTextFiles();
	}
        vikSleep(1 + random() % (random() % 15 + 1));
        if ((random() & 7) == 0)
            vikOut("say %s", txtRandom(saying, nSayings));
        else if ((random() & 3) == 0)
            vikOut("%s %s",
                   txtRandom(feeling, nFeelings),
                   txtRandom(adverb, nAdverbs));
        vikOut("get all");
        vikOut("drop corpse");
	if (invWhen-- == 0) {
	    vikOut("i");
	    invWhen = 5;
	}
	if (scoreWhen-- == 0) {
            vikOut("score -l");
	    scoreWhen = 30;
	}
	if (wearWhen-- == 0) {
            vikOut("unequip");
            vikOut("wield knife");
            vikOut("wield mace");
            vikOut("wield hammer");
            vikOut("wield sword");
            vikOut("wear platemail");
            vikOut("wear all");
	    wearWhen = 10;
	}
	if (eatWhen-- == 0) {
            vikOut("eat all");
            vikOut("drink all");
	    eatWhen = 25;
	}
	if (timeWhen-- == 0) {
	    struct tm *tm;
	    time_t t;
	    char   tmp[20];
	    
	    time(&t);
	    tm = localtime(&t);
	    strftime(tmp, 20, "%x %X", tm);
 	    msgVerbose(1, "***  Current time: %s\n", tmp);
	    msgVerbose(1, "***  Status: HP=%d, SP=%d, XP=%d, "
		       "XPNL=%d, CASH=%d\n",
		       hp, sp, xp, xpnl, cash);
	    timeWhen = 5;
	}
        vikSleep(2);
        vikOut("glance");
        while (vikWaitForTimeout("[", 2000) != 0) {
	    msgVerbose(1, "***  timing out for new direction. "
		       "hyperspacing!\n");
	    /* flee didn't always work */
	    if ((random() % NUM_DIRECTIONS) == 0) {
		vikOut("flee");
		*prevExit = '\0';
	    } else {
		strcpy(prevExit, direction[random() % NUM_DIRECTIONS]);
		vikOut(prevExit);
	    }
	    vikOut("glance");
	}
        vikIn(line, sizeof(line));
        if ((s = strchr(line, ']')) != NULL) {
            *s = '\0';
            tok = strtok(line, " ");
            nExit = 0;
            *str = '\0';
            while (tok) {
                exits[nExit++] = tok;
                if (*str)
                    strcat(str, ", ");
                strcat(str, tok);
                tok = strtok(NULL, " ");
            }
	    /* check the creator before doing a new move. */
	    if (!checkCreator()) {
		do {
		    n = random() % nExit;
		} while (nExit > 1
			 && strcmp(exits[n], oppositeDirection(prevExit)) == 0
			 && (random() % 10) != 0
			 );
		msgVerbose(1, "***  choosing %s from { %s }\n", exits[n], str);
		strncpy(prevExit, exits[n], sizeof(prevExit) - 1);
		vikOut(exits[n]);
	    }
        }
    }
}



/**************************************************************************
 *                                                                        *
 *                    P U B L I C    F U N C T I O N S                    *
 *                                                                        *
 **************************************************************************/

int main(int argc, char *argv[])
{
    msgSetName(argv[0]);
    msgSetVerbose(10);
    srandom(time(NULL) + getpid());
    readTextFiles();
    vikLogin(PLAYER_NAME, PLAYER_PASS);
    setupSigHandlers();
    robot();
    vikLogout();
    return 0;
}
